import pyranges as pr

def _drop_duplicate_positions(df, kwargs):

    strand = kwargs.get("strand")

    columns = ["Start", "End"]
    if strand:
        columns.append("Strand")

    return df.drop_duplicates(columns)



import pandas as pd
import numpy as np

from pyranges.subset import (get_string, get_tuple, get_slice, get_booldict)
from pyranges.methods.drop import _drop, _keep

from pyranges import PyRanges


def _getitem(self, val):

    if isinstance(val, list):
        dfs = _keep(self, keep=val).dfs
    elif isinstance(val, str):
        dfs = get_string(self, val)
    elif isinstance(val, tuple):
        dfs = get_tuple(self, val)
    elif isinstance(val, slice):
        dfs = get_slice(self, val)
    elif isinstance(val, dict):
        dfs = get_booldict(self, val)
    elif (isinstance(val, (pd.Series, np.ndarray))) and val.dtype == "bool":
        assert len(val) == len(self), "Boolean indexer must be same length as pyrange!"
        grpby = "Chromosome" if not self.stranded else ["Chromosome", "Strand"]
        to_grpby = ["Chromosome"] if not self.stranded else [self.Chromosome, self.Strand]
        d = {k: v.iloc[:, -1] for k, v in pd.concat(to_grpby + [pd.Series(val)], axis=1).groupby(grpby)}
        dfs = get_booldict(self, d)
    else:
        raise Exception("Not valid subsetter: {}".format(str(val)))

    gr = PyRanges(dfs)
    return gr



import pytest
from natsort import natsorted

from hypothesis import given, settings, HealthCheck  #, assume
from hypothesis import reproduce_failure  # pylint: disable=unused-import

import tempfile
import subprocess  # nosec
from io import StringIO

import numpy as np

import pandas as pd

from tests.helpers import assert_df_equal
from tests.hypothesis_helper import dfs_min, df_data, selector, dfs_min_with_id, max_examples, deadline

import pyranges as pr

from os import environ

# if environ.get("TRAVIS"):
#     max_examples = 100
#     deadline = None
# else:
#     max_examples = 1000
#     deadline = None

merge_command = "bedtools merge -o first -c 6 {} -i <(sort -k1,1 -k2,2n {})"


@pytest.mark.bedtools
@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_merge(gr, strand):

    bedtools_strand = {True: "-s", False: ""}[strand]

    print(gr)

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = merge_command.format(bedtools_strand, f1)
        print(cmd)

        # ignoring bandit security warning. All strings created by test suite
        result = subprocess.check_output(  # nosec
            cmd, shell=True, executable="/bin/bash").decode()  # nosec

        if not strand:
            bedtools_df = pd.read_csv(
                StringIO(result),
                sep="\t",
                header=None,
                squeeze=True,
                usecols=[0, 1, 2],
                names="Chromosome Start End".split(),
                dtype={"Chromosome": "category"})
        else:
            bedtools_df = pd.read_csv(
                StringIO(result),
                sep="\t",
                header=None,
                squeeze=True,
                names="Chromosome Start End Strand".split(),
                dtype={"Chromosome": "category"})

    print("bedtools_df\n", bedtools_df)
    result = gr.merge(strand=strand)
    print("result\n", result.df)

    if not bedtools_df.empty:
        # need to sort because bedtools sometimes gives the result in non-natsorted chromosome order!
        if result.stranded:
            assert_df_equal(
                result.df.sort_values("Chromosome Start Strand".split()),
                bedtools_df.sort_values("Chromosome Start Strand".split()))
        else:
            assert_df_equal(
                result.df.sort_values("Chromosome Start".split()),
                bedtools_df.sort_values("Chromosome Start".split()))
    else:
        assert bedtools_df.empty == result.df.empty


cluster_command = "bedtools cluster {} -i <(sort -k1,1 -k2,2n {})"



@pytest.mark.bedtools
@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_cluster(gr, strand):

    bedtools_strand = {True: "-s", False: ""}[strand]

    print(gr)

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = cluster_command.format(bedtools_strand, f1)
        print(cmd)

        # ignoring bandit security warning. All strings created by test suite
        result = subprocess.check_output(  # nosec
            cmd, shell=True, executable="/bin/bash").decode()  # nosec

        bedtools_df = pd.read_csv(
            StringIO(result),
            sep="\t",
            header=None,
            squeeze=True,
            names="Chromosome Start End Name Score Strand Cluster".split(),
            dtype={"Chromosome": "category"})

    print("bedtools_df\n", bedtools_df)

    # from pydbg import dbg
    # dbg(gr.cluster(strand=strand))

    print("gr\n", gr)
    result = gr.cluster(strand=strand)
    print("result\n", result[["Cluster"]])
    print("result\n", result.df)

    if not bedtools_df.empty:
        # need to sort because bedtools sometimes gives the result in non-natsorted chromosome order!
        if result.stranded:
            sort_values = "Chromosome Start Strand".split()
        else:
            sort_values = "Chromosome Start".split()

        result_df = result.df.sort_values(sort_values)
        bedtools_df = bedtools_df.sort_values(sort_values)

        cluster_ids = {
            k: v
            for k, v in zip(result_df.Cluster.drop_duplicates(),
                            bedtools_df.Cluster.drop_duplicates())
        }

        # bedtools gives different cluster ids than pyranges
        result_df.Cluster.replace(cluster_ids, inplace=True)
        assert_df_equal(result_df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min_with_id())  # pylint: disable=no-value-for-parameter
def test_cluster_by(gr, strand):

    result = gr.cluster(by="ID", strand=strand).df
    print(result)
    df = gr.df

    if strand:
        groupby = ["Chromosome", "Strand", "ID"]
    else:
        groupby = ["Chromosome", "ID"]


    grs = []

    for g, gdf in natsorted(df.groupby(groupby)):
        grs.append(pr.PyRanges(gdf))


    clusters = [gr.cluster(strand=strand) for gr in grs]
    i = 1
    new_clusters = []
    for c in clusters:
        print("c")
        print(c)
        c.Cluster = i
        i += 1
        new_clusters.append(c)

    expected = pr.concat(new_clusters).df
    expected.loc[:, "Cluster"] = expected.Cluster.astype(np.int32)
    # expected = expected.drop_duplicates()

    print(expected)
    print(result)

    assert_df_equal(result, expected)


@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min_with_id())  # pylint: disable=no-value-for-parameter
def test_merge_by(gr, strand):

    print(gr)
    result = gr.merge(by="ID").df.drop("ID", axis=1)

    df = gr.df

    grs = []
    for g, gdf in df.groupby("ID"):
        grs.append(pr.PyRanges(gdf))

    expected = pr.concat([gr.merge() for gr in grs]).df

    print(expected)
    print(result)

    assert_df_equal(result, expected)




makewindows_command = "bedtools makewindows -w 10 -b <(sort -k1,1 -k2,2n {})"


@pytest.mark.bedtools
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_windows(gr):

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = makewindows_command.format(f1)
        print(cmd)

        # ignoring bandit security warning. All strings created by test suite
        result = subprocess.check_output(  # nosec
            cmd, shell=True, executable="/bin/bash").decode()  # nosec

        bedtools_df = pd.read_csv(
            StringIO(result),
            sep="\t",
            header=None,
            squeeze=True,
            names="Chromosome Start End".split(),
            dtype={"Chromosome": "category"})

    print("bedtools_df\n", bedtools_df)

    result = gr.window(10)["Chromosome Start End".split()].unstrand()
    print("result\n", result.df)

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=df_data())  # pylint: disable=no-value-for-parameter
def test_init(gr, strand):

    c, s, e, strands = gr

    if strand:
        pr.PyRanges(chromosomes=c, starts=s, ends=e, strands=strands)
    else:
        pr.PyRanges(chromosomes=c, starts=s, ends=e)


chipseq = pr.data.chipseq()


@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(selector=selector())  # pylint: disable=no-value-for-parameter
def test_getitem(selector):

    # have these weird returns to avoid being flagged as unused code
    if len(selector) == 3:
        a, b, c = selector
        return chipseq[a, b, c]
    elif len(selector) == 2:
        a, b = selector
        return chipseq[a, b]
    elif len(selector) == 1:
        a = selector[0]
        return chipseq[a]
    elif len(selector) == 0:
        pass
    else:
        raise Exception("Should never happen")


@pytest.mark.bedtools
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_summary(gr):

    # merely testing that it does not error
    # contents are just (pandas) dataframe.describe()
    gr.summary()



from pyranges.methods.join import _both_dfs

import numpy as np
import pandas as pd

def _insert(scdf, ocdf, kwargs):

    how = kwargs.get("how")
    col = kwargs.get("columns")
    suffix = kwargs.get("suffix")
    overlap_only = kwargs.get("overlap_only")

    if scdf.empty or ocdf.empty:
        return None

    _scdf, ocdf = _both_dfs(scdf, ocdf, how=how)

    nix = pd.Index(range(len(_scdf)))
    _scdf.index = nix
    ocdf.index = nix

    ocdf = ocdf[col]

    df = _scdf.join(ocdf, rsuffix=suffix)

    if not overlap_only:
        # need to insert nan for those rows in self which had no overlap

        scdf = scdf[~scdf.index.isin(_scdf.index)]

        if isinstance(col, str):
            scdf.insert(scdf.shape[1], col, np.nan)
        else:
            for c in col:
                scdf.insert(scdf.shape[1], c, np.nan)

        df = pd.concat([df, scdf])

    df = df.sort_values("Start End".split())

    return df



import re
from collections.abc import Iterable


def _drop(self, drop=None):
    columns = self.columns
    if "Strand" in columns:
        self = self.unstrand()
        columns = [c for c in columns if c != "Strand"]

    if self.stranded:
        always_keep = "Chromosome Start End Strand".split()
    else:
        always_keep = "Chromosome Start End".split()

    _to_drop = []

    if not drop:
        _to_drop = set(columns) - set(always_keep)
    elif isinstance(drop, str):
        _to_drop = [drop]
    elif isinstance(drop, Iterable) or isinstance(drop, list):
        _to_drop = drop
    else:
        raise Exception("Not valid subsetters!")

    _to_drop = set(_to_drop) - set(always_keep)

    return self.apply(lambda df: df.drop(_to_drop, axis=1))


def _keep(self, keep):

    columns = self.columns
    if not self.stranded:
        always_keep = "Chromosome Start End".split()
    else:
        always_keep = "Chromosome Start End Strand".split()

    _to_drop = []
    if isinstance(keep, Iterable) or isinstance(keep, list):
        _to_drop = set(columns) - set(keep) - set(always_keep)
    else:
        raise Exception("Column(s) to drop must be in list.")

    self = self.apply(lambda df: df.drop(_to_drop, axis=1))

    columns = self.columns

    keep = [c for c in keep if not c in always_keep]
    new_order = []
    i = 0
    for c in columns:
        if c in always_keep:
            new_order.append(c)
        else:
            new_order.append(keep[i])
            i += 1

    self = self.apply(lambda df: df[new_order])

    return self



from io import StringIO
import pandas as pd
import pyranges as pr
import pytest

from hypothesis import given, settings, HealthCheck  #, assume
from hypothesis import reproduce_failure  # pylint: disable=unused-import

from tests.hypothesis_helper import genomicfeature

from tests.hypothesis_helper import max_examples, deadline, slow_max_examples

by_to_id = {"gene": "gene_id", "transcript": "transcript_id"}

def compute_introns_single(df, by):
    id_column = by_to_id[by]
    g = df[df.Feature == by]
    x = pr.PyRanges(df[df.Feature == "exon"]).merge(by=id_column).df

    x = pd.concat([x, x.tail(1)], sort=False).reset_index(drop=True)
    cols = list(x.columns)
    original_cols = cols[::1]
    cols[2] = "Start"
    cols[1] = "End"
    x = x[cols]
    x.columns = original_cols
    x.loc[:, "Start"] = x.Start.shift()
    x.at[x.index[0], "Start"] = g.Start.iloc[0]
    x.at[x.index[-1], "End"] = g.End.iloc[0]
    x.loc[:, "Start"] = x.Start.astype(int)
    x = x[x.Start != x.End]

    return x


# def test_introns_single():

#     "Assert that our fast method of computing introns is the same as the slow, correct one in compute_introns_single"

#     gr = pr.data.gencode_gtf()[["gene_id", "Feature"]]
#     exons = gr[gr.Feature == "exon"].merge(by="gene_id")
#     exons.Feature = "exon"
#     exons = exons.df
#     df = pd.concat([gr[gr.Feature == "gene"].df, exons], sort=False)
#     print(df)

#     for gid, gdf in df.groupby("gene_id"):
#         print("-------" * 20)
#         print(gid)
#         print(gdf)
#         print("gdf", len(gdf))
#         expected = compute_introns_single(gdf)
#         print("expected", len(expected))
#         actual = pr.PyRanges(gdf).features.introns().df
#         print("actual", len(actual))
#         if actual.empty:
#             assert expected.empty
#             continue

#         assert list(expected.Start) == list(actual.Start)
#         assert list(expected.End) == list(actual.End)



def _introns_correct(full, genes, exons, introns, by):

    """Testing that introns:

    1: ends larger than starts
    2: the intersection of the computed introns and exons per gene are 0
    3: that the number of introns overlapping each gene is the same as number of introns per gene
    4 & 5: that the intron positions are the same as the ones computed with the slow, but correct algo"""
    id_column = by_to_id[by]
    if len(introns):
        assert (introns.Start < introns.End).all(), str(introns[(introns.Start >= introns.End)])

    expected_results = {}
    based_on = {}
    for gene_id, gdf in full.groupby(id_column): # #[full.gene_id.isin(["ENSG00000078808.16"])]
        expected_results[gene_id] = compute_introns_single(gdf, by)
        based_on[gene_id] = pr.PyRanges(gdf[gdf.Feature.isin([by, "exon"])]).df

    if not len(introns):
        for v in expected_results.values():
            assert v.empty
        return "okay"

    for gene_id, idf in introns.groupby(id_column):
        expected = expected_results[gene_id]
        # print("based_on", based_on[gene_id])
        exons = pr.PyRanges(based_on[gene_id]).subset(lambda df: df.Feature == "exon").merge(by=id_column)
        genes = pr.PyRanges(based_on[gene_id]).subset(lambda df: df.Feature == by)
        # print("exons", exons)
        # print("actual", idf)
        # print("expected", expected)
        _introns = pr.PyRanges(idf)
        assert len(exons.intersect(_introns)) == 0
        assert len(genes.intersect(_introns)) == len(_introns)
        assert list(idf.Start) == list(expected.Start), "not equal"
        assert list(idf.End) == list(expected.End), "not equal"

by = ["gene", "transcript"]



@pytest.mark.parametrize("by", by)
@settings(
    max_examples=slow_max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=genomicfeature())  # pylint: disable=no-value-for-parameter
def test_introns(gr, by):

    id_column = by_to_id[by]

    delim = "-" * 20
    print("\n" + delim +" new test " + delim)
    print(gr)

    genes = gr[gr.Feature.isin(["gene"])].df
    exons = gr[gr.Feature.isin(["exon"])].merge(by=id_column).df
    introns = gr.features.introns(by=by).df

    _introns_correct(gr.df, genes, exons, introns, by)

# @pytest.mark.parametrize("by", ["transcript"])
# def test_introns(by):

#     id_column = by_to_id[by]

#     gr = pr.data.gencode_gtf()[[id_column, "Feature"]].sort(id_column)
#     genes = gr[gr.Feature.isin(["gene"])].df
#     exons = gr[gr.Feature.isin(["exon"])].merge(by=id_column).df
#     introns = gr.features.introns(by=by).df

#     _introns_correct(gr.df, genes, exons, introns, by)



# @pytest.mark.parametrize("by", by)
# def test_introns_gencode(by):

#     gr = pr.data.gencode_gtf()
#     introns = gr.features.introns(by=by)


# @pytest.mark.parametrize("by", by)
# def test_introns_ensembl(by):

#     gr = pr.data.ensembl_gtf()
#     introns = gr.features.introns(by=by)

# @pytest.mark.parametrize("by", by)
# def test_introns_ucsc(by):

#     gr = pr.data.ucsc_bed()
#     introns = gr.features.introns(by=by)


# @pytest.fixture
# def simple_exons():

#     c = """Chromosome    Start      End Strand             gene_id Feature
# chr1  1471769  1472089      +  ENSG00000160072.19 exon
# chr1  1477274  1477350      +  ENSG00000160072.19 exon
# chr1  1478026  1478745      +  ENSG00000160072.19 exon
# chr1  1479049  1479108      +  ENSG00000160072.19 exon
# chr1  1480867  1480936      +  ENSG00000160072.19 exon
# chr1  1482138  1482303      +  ENSG00000160072.19 exon
# chr1  1482545  1482662      +  ENSG00000160072.19 exon
# chr1  1483485  1485171      +  ENSG00000160072.19 exon
# chr1  1485782  1486235      +  ENSG00000160072.19 exon
# chr1  1486544  1486668      +  ENSG00000160072.19 exon
# chr1  1487863  1487914      +  ENSG00000160072.19 exon
# chr1  1489204  1490671      +  ENSG00000160072.19 exon
# chr1  1495485  1497848      +  ENSG00000160072.19 exon"""

#     return pr.PyRanges(pd.read_csv(StringIO(c), sep=r"\s+"))


# @pytest.fixture
# def simple_gene():

#     c = """Chromosome    Start      End Strand             gene_id Feature
# chr1    1471769  1497848      +  ENSG00000160072.19  gene"""

#     return pr.PyRanges(pd.read_csv(StringIO(c), sep=r"\s+"))


# def test_simple_introns(simple_exons, simple_gene):

#     gr = pr.concat([simple_exons, simple_gene])
#     introns = gr.features.introns()
#     expected_starts = [1472089, 1477350, 1478745, 1479108, 1480936, 1482303, 1482662, 1485171, 1486235, 1486668, 1487914, 1490671]
#     expected_ends = [1477274, 1478026, 1479049, 1480867, 1482138, 1482545, 1483485, 1485782, 1486544, 1487863, 1489204, 1495485]
#     assert list(introns.Start) == expected_starts
#     assert list(introns.End) ==  expected_ends
#     assert len(introns.intersect(simple_exons)) == 0



from hypothesis.extra.pandas import data_frames, column, indexes
from hypothesis.extra.numpy import arrays
import hypothesis.strategies as st

import pyranges as pr
from pyranges import PyRanges

import pandas as pd
import numpy as np

from os import environ

if environ.get("TRAVIS"):
    max_examples = 100
    slow_max_examples = 10
    deadline = None
else:
    max_examples = 1000
    slow_max_examples = 100
    deadline = None

lengths = st.integers(min_value=1, max_value=int(1e7))
small_lengths = st.integers(min_value=1, max_value=int(1e4))

strands = st.sampled_from("+ -".split())
single_strand = st.sampled_from(["+"])
names = st.text("abcdefghijklmnopqrstuvxyz", min_size=1)
scores = st.integers(min_value=0, max_value=256)

datatype = st.sampled_from([pd.Series, np.array, list])

feature_data = st.sampled_from(["ensembl_gtf", "gencode_gtf", "ucsc_bed"])

chromosomes = st.sampled_from(
    ["chr{}".format(str(e)) for e in list(range(1, 23)) + "X Y M".split()])
chromosomes_small = st.sampled_from(["chr1"])
cs = st.one_of(chromosomes, chromosomes_small)

runlengths = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True),
    columns=[
        column("Runs", st.integers(min_value=1, max_value=int(1e7))),
        # must have a min/max on floats because R S4vectors translates too big ones into inf.
        # which is unequal to eg -1.79769e+308 so the tests fail
        column("Values", st.integers(min_value=-int(1e7), max_value=int(1e7)))
    ])

better_dfs_no_min = data_frames(
    index=indexes(dtype=np.int64, min_size=0, unique=True, elements=lengths),
    columns=[
        column("Chromosome", cs),
        column("Start", elements=lengths),
        column("End", elements=small_lengths),
        # column("Name", elements=names),
        # column("Score", elements=scores),
        column("Strand", strands)
    ])

better_dfs_min = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True, elements=lengths),
    columns=[
        column("Chromosome", cs),
        column("Start", elements=lengths),
        column("End", elements=small_lengths),
        # column("Name", elements=names),
        # column("Score", elements=scores),
        column("Strand", strands)
    ])

better_dfs_min_2 = data_frames(
    index=indexes(dtype=np.int64, min_size=2, unique=True, elements=lengths),
    columns=[
        column("Chromosome", chromosomes_small),
        column("Start", elements=lengths),
        column("End", elements=small_lengths),
        # column("Name", elements=names),
        # column("Score", elements=scores),
        column("Strand", single_strand)
    ])

better_dfs_min_single_chromosome = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True, elements=lengths),
    columns=[
        column("Chromosome", chromosomes_small),
        column("Start", elements=lengths),
        column("End", elements=small_lengths),
        # column("Name", elements=names),
        # column("Score", elements=scores),
        column("Strand", strands)
    ])

runlengths_same_length_integers = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True),
    columns=[
        column("Runs", st.integers(min_value=1, max_value=int(1e4))),
        column("Values", st.integers(min_value=1, max_value=int(1e4))),
        column("Values2", st.integers(min_value=1, max_value=int(1e4)))
    ])


@st.composite
def dfs_min2(draw):  # nosec
    df = draw(better_dfs_min_2)
    # strand = draw(use_strand)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    # if not strand:
    #     df = df.drop("Strand", axis=1)

    gr = PyRanges(df, int64=True)
    # gr = PyRanges(df)

    # do not sort like this, use pyranges sort
    # np.random.seed(draw(st.integers(min_value=0, max_value=int(1e6))))
    # gr.df = df.reindex(np.random.permutation(df.index.values))

    return gr


@st.composite
def dfs_min(draw):  # nosec
    df = draw(better_dfs_min)
    # strand = draw(use_strand)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    # df.Start = df.Start.astype(np.int32)
    # df.End = df.End.astype(np.int32)
    # print(df.dtypes)
    # stranded = draw(st.booleans())
    # if not strand:
    #     df = df.drop("Strand", axis=1)

    gr = PyRanges(df, int64=True)
    # print(gr)
    # raise
    # gr = PyRanges(df)

    # do not sort like this, use pyranges sort
    # np.random.seed(draw(st.integers(min_value=0, max_value=int(1e6))))
    # gr.df = df.reindex(np.random.permutation(df.index.values))

    return gr


@st.composite
def dfs_no_min(draw):  # nosec
    df = draw(better_dfs_no_min)
    # strand = draw(use_strand)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    # stranded = draw(st.booleans())
    # if not strand:
    #     df = df.drop("Strand", axis=1)

    gr = PyRanges(df, int64=True)
    # gr = PyRanges(df)

    # do not sort like this, use pyranges sort
    # np.random.seed(draw(st.integers(min_value=0, max_value=int(1e6))))
    # gr.df = df.reindex(np.random.permutation(df.index.values))

    return gr


@st.composite
def dfs_min_with_id(draw):  # nosec
    df = draw(better_dfs_min)
    ids = df.Start
    # strand = draw(use_strand)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)
    df.insert(5, "ID", ids)

    # df.Start = df.Start.astype(np.int32)
    # df.End = df.End.astype(np.int32)
    # print(df.dtypes)
    # stranded = draw(st.booleans())
    # if not strand:
    #     df = df.drop("Strand", axis=1)

    gr = PyRanges(df)
    # print(gr)
    # raise
    # gr = PyRanges(df)

    # do not sort like this, use pyranges sort
    # np.random.seed(draw(st.integers(min_value=0, max_value=int(1e6))))
    # gr.df = df.reindex(np.random.permutation(df.index.values))

    return gr


@st.composite
def dfs_min_with_gene_id(draw):  # nosec
    df = draw(better_dfs_min)
    ids = df.Start
    # strand = draw(use_strand)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)
    df.insert(5, "gene_id", ids)
    df.insert(6, "exon_id", list(range(len(df))))

    # df.Start = df.Start.astype(np.int32)
    # df.End = df.End.astype(np.int32)
    # print(df.dtypes)
    # stranded = draw(st.booleans())
    # if not strand:
    #     df = df.drop("Strand", axis=1)

    gr = PyRanges(df)
    # print(gr)
    # raise
    # gr = PyRanges(df)

    # do not sort like this, use pyranges sort
    # np.random.seed(draw(st.integers(min_value=0, max_value=int(1e6))))
    # gr.df = df.reindex(np.random.permutation(df.index.values))

    return gr


@st.composite
def df_data(draw):

    df = draw(better_dfs_min)
    df.loc[:, "End"] += df.Start

    chromosome_type = draw(datatype)
    start_type = draw(datatype)
    end_type = draw(datatype)
    strand_type = draw(datatype)

    strand = strand_type(df.Strand)
    chromosome = chromosome_type(df.Chromosome)
    start = start_type(df.Start)
    end = end_type(df.End)

    return chromosome, start, end, strand


@st.composite
def dfs_min_single_chromosome(draw):
    df = draw(better_dfs_min_single_chromosome)
    df.loc[:, "End"] += df.Start
    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    return df


@st.composite
def genomicfeature(draw):

    dataset_name = draw(feature_data)
    print("dataset name " * 5, dataset_name)
    dataset = getattr(pr.data, dataset_name)()
    dataset = dataset[dataset.Feature.isin(["gene", "transcript", "exon"])]

    # subsetter = draw(arrays(np.bool, shape=len(dataset)))
    length = draw(st.integers())
    gene_ids = list(set(dataset.gene_id))
    genes = draw(st.lists(st.sampled_from(gene_ids), unique="True", min_size=1))
    dataset = dataset[dataset.gene_id.isin(genes)]

    return dataset





@st.composite
def selector(draw):

    df = draw(better_dfs_min)
    h = df.head(1)
    chromosome = h["Chromosome"].iloc[0]
    start = h["Start"].iloc[0]
    end = h["End"].iloc[0]
    strand = h["Strand"].iloc[0]

    chromosome_bool = draw(st.booleans())
    strand_bool = draw(st.booleans())
    start_bool = draw(st.booleans())
    end_bool = draw(st.booleans())

    _slice = {
        (True, True): slice(start, end),
        (True, False): slice(start, None),
        (False, True): slice(None, end),
        (False, False): slice(None, None)
    }[start_bool, end_bool]

    to_return = []
    if chromosome_bool:
        to_return.append(chromosome)
    if strand_bool:
        to_return.append(strand)
    if start_bool or end_bool:
        to_return.append(_slice)

    return to_return



import pandas as pd

from sorted_nearest import find_clusters, merge_by


def _merge(df, kwargs):

    if df.empty:
        return None

    slack = kwargs.get("slack", 0)
    chromosome, strand = kwargs["chromosome"], kwargs.get("strand", None)

    cdf = df.sort_values("Start")

    starts, ends = find_clusters(cdf.Start.values, cdf.End.values, slack)

    nidx = pd.Index(range(len(starts)))
    if strand:
        cluster_df = pd.DataFrame({
            "Chromosome":
            pd.Series(chromosome, dtype="category", index=nidx),
            "Start":
            starts,
            "End":
            ends,
            "Strand":
            pd.Series(strand, dtype="category", index=nidx)
        })
    else:
        cluster_df = pd.DataFrame({
            "Chromosome":
            pd.Series(chromosome, dtype="category", index=nidx),
            "Start":
            starts,
            "End":
            ends
        })

    return cluster_df


def _merge_by(df, kwargs):

    if df.empty:
        return None

    slack = kwargs.get("slack", 0)
    by = kwargs["by"]
    chromosome, strand = kwargs["chromosome"], kwargs.get("strand", None)

    cdf = df.sort_values([by, "Start"])

    new_ids = (cdf[by] != cdf[by].shift()).cumsum()

    new_to_old = pd.DataFrame(data={"new": new_ids.drop_duplicates(), "old": cdf[by].drop_duplicates()})

    cdf.insert(cdf.shape[1], "ClusterBy", new_ids)

    ids, starts, ends = merge_by(cdf.Start.values, cdf.End.values, cdf.ClusterBy.values, slack)

    nidx = pd.Index(range(len(starts)))
    if strand:
        cluster_df = pd.DataFrame({
            "Chromosome":
            pd.Series(chromosome, dtype="category", index=nidx),
            "Start":
            starts,
            "End":
            ends,
            by:
            ids,
            "Strand":
            pd.Series(strand, dtype="category", index=nidx)
        })
    else:
        cluster_df = pd.DataFrame({
            "Chromosome":
            pd.Series(chromosome, dtype="category", index=nidx),
            "Start":
            starts,
            "End":
            ends,
            by:
            ids,
        })

    cluster_df = cluster_df.merge(new_to_old, left_on=by, right_on="new")
    cluster_df = cluster_df.drop([by, "new"], axis=1)
    cluster_df = cluster_df.rename(columns={"old": by})

    return cluster_df



import pyranges as pr
import pandas as pd

from collections import defaultdict


def concat(pyranges, strand=False):

    # from pydbg import dbg
    pyranges = [pr for pr in pyranges if not pr.empty]
    # dbg(pyranges)
    # dbg([p.df.dtypes for p in pyranges])
    grs_per_chromosome = defaultdict(list)

    if strand:
        assert all([
            gr.stranded for gr in pyranges
        ]), "Cannot do stranded concat, not all pyranges contain strand info."

        for gr in pyranges:
            for k, df in gr.dfs.items():
                # dbg(df)
                grs_per_chromosome[k].append(df)
    else:
        for gr in pyranges:
            for chromosome in gr.chromosomes:
                # dbg(gr)
                # dbg(gr[chromosome])
                df = gr[chromosome].df
                # dbg(df.dtypes)
                grs_per_chromosome[chromosome].append(df)

    new_pyrange = {}

    for k, v in grs_per_chromosome.items():
        # dbg([_v.dtypes for _v in v])
        new_pyrange[k] = pd.concat(v, sort=False)
        # dbg(new_pyrange[k].dtypes)

    res = pr.multithreaded.process_results(new_pyrange.values(),
                                           new_pyrange.keys())

    # dbg([r.dtypes for r in res.values()])

    return pr.PyRanges(res)



from collections import OrderedDict

import pandas as pd

from tabulate import tabulate


def _summary(self):

    lengths = OrderedDict()
    lengths["pyrange"] = self.lengths()

    if self.stranded:
        c = self.merge(strand=True)
        lengths["coverage_stranded"] = c.lengths()

    c = self.merge(strand=False)
    lengths["coverage_unstranded"] = c.lengths()

    summaries = OrderedDict()

    for summary, d in lengths.items():
        summaries[summary] = pd.concat(d.values()).describe()

    summary = pd.concat(summaries.values(), axis=1)
    summary.columns = list(summaries)

    str_repr = tabulate(summary, headers=summary.columns, tablefmt='psql')
    print(str_repr)
